"""
Module for post-processing and analyzing simulation results.
"""

import numpy as np
import pandas as pd
import allel
from typing import Tuple, Optional, Dict
import itertools
from math import radians
from sklearn.metrics.pairwise import haversine_distances
from sklearn import linear_model
from esda.moran import Moran
from libpysal.weights import Voronoi
from scipy.stats import entropy


def filter_gt(
    gt: np.ndarray,
    deme_dict_inds: dict = None,
    deme_dict_anc: dict = None,
    missing_data_perc: float = 0,
    r2_thresh: float = 0.1,
    filter_monomorphic: bool = True,
    filter_singletons: bool = True,  # Added new parameter with default value True
) -> Tuple[
    allel.GenotypeArray,
    allel.AlleleCountsArray,
    Optional[Dict[str, allel.AlleleCountsArray]],
    Optional[Dict[str, allel.AlleleCountsArray]],
]:
    """
    Filter genotype matrices output by ts.genotype_matrix() to filter out monomorphic sites, loci in linkage disequilibrium, and recreate missing data patterns common to empirical genotype data.
    Returns the genotype matrix and allele counts matrix for the filtered loci, and optionally allele counts matrices for demes and ancestral populations.

    Parameters
    ----------
    gt : np.ndarray
        The genotype matrix.
    deme_dict_inds : dict, optional
        A dictionary containing the indices of individuals in each deme. Defaults to None.
    deme_dict_anc : dict, optional
        A dictionary containing the indices of individuals in each ancestral population. Defaults to None.
    missing_data_perc : float, optional
        The percentage of missing data allowed. Defaults to 0.
    r2_thresh : float, optional
        The threshold for linkage disequilibrium. Defaults to 0.1.
    filter_monomorphic : bool, optional
        Whether to filter out monomorphic sites, keeping only segregating sites. Defaults to True.
    filter_singletons : bool, optional
        Whether to filter out singletons. Defaults to True.

    Returns
    -------
    Tuple[allel.GenotypeArray, allel.AlleleCountsArray, Optional[Dict[str, allel.AlleleCountsArray]], Optional[Dict[str, allel.AlleleCountsArray]]]
        A tuple containing the filtered genotype matrix, the allele counts matrix, a dictionary of allele counts matrices for demes (if deme_dict_inds is provided), and a dictionary of allele counts matrices for ancestral populations (if deme_dict_anc is provided).

    Notes
    -----
    This function uses a random mask to simulate missing data in the genotype matrix.
    For reproducibility it's advised to set a `np.random.seed()` before calling this function.
    """

    # create data mask that masks the tree sequence genotype matrix to have similar missing data patterns to the empirical data

    # get total the number of elements in the genotype matrix
    total_array_size = gt.shape[0] * gt.shape[1]

    # fill an array with zeros that is the total size of the original matrix
    a = np.zeros(total_array_size, dtype=int)

    # set the first x elements to be the size of non-missing data
    non_miss = int(np.ceil(total_array_size * (1 - missing_data_perc)))

    # 1 means "True"
    a[:non_miss] = 1

    # randomly shuffle True and False values
    np.random.shuffle(a)

    # transform to boolean
    a = a.astype(bool)

    # reshape to genotype matrix shape
    miss_mask = np.reshape(a, gt.shape)

    # mask the genotype matrix
    geno_mat = np.ma.masked_array(gt, mask=miss_mask, fill_value=-1)

    # convert to a GenotypeArray
    h = allel.HaplotypeArray(geno_mat)

    gt = h.to_genotypes(ploidy=2)

    # get the allele counts matrix for all individuals
    ac = gt.count_alleles(max_allele=1)

    if filter_monomorphic:
        # only retain monomorphic alleles
        is_seg = ac.is_segregating()
        gt = gt.compress(is_seg, axis=0)
        ac_seg = ac.compress(is_seg)
    else:
        ac_seg = ac

    # alt allele counts matrix
    # needed for finding unlinked SNPs
    gn = gt.to_n_alt(fill=-1)

    # filter for unlinked loci
    # locate unlinked SNPs
    loc_unlinked = allel.locate_unlinked(gn, threshold=r2_thresh)

    # select unlinked SNPs
    ac_unlinked = ac_seg[loc_unlinked]
    gt_unlinked = gt[loc_unlinked]

    if filter_singletons:
        # remove singletons
        not_singleton = ~ac_unlinked.is_singleton(allele=1)
        ac_unlinked = ac_unlinked[not_singleton]
        gt_unlinked = gt_unlinked[not_singleton]

    if deme_dict_inds is not None:
        ac_unlinked_demes = gt_unlinked.count_alleles_subpops(
            max_allele=1, subpops=deme_dict_inds
        )
    else:
        ac_unlinked_demes = None

    # create allele counts matrices for the ancestral populations
    if deme_dict_anc is not None:
        anc_dict_inds = {}
        for key in deme_dict_anc:
            demes = deme_dict_anc[key]
            anc_dict_inds[key] = []
            for d in demes:
                # convert numpy array to list
                inds = deme_dict_inds[d].tolist()
                anc_dict_inds[key].extend(inds)

        ac_unlinked_anc = gt_unlinked.count_alleles_subpops(
            max_allele=1, subpops=anc_dict_inds
        )

    else:
        ac_unlinked_anc = None

    return gt_unlinked, ac_unlinked, ac_unlinked_demes, ac_unlinked_anc


def calc_sumstats(
    ac: allel.AlleleCountsArray,
    coords_dict: dict,
    anc_demes_dict: dict = None,
    ac_demes: dict = None,
    ac_anc: dict = None,
    between_anc_pop_sumstats: bool = False,
    return_df: bool = False,
    precision: int = 6,
) -> dict:
    """
    Calculates a suite of genetic summary statistics on allele counts matrices generated by `filter_gt` or otherwise generated through scikit-allel.
    The required input is an allele counts matrix for all individuals/demes, and a dictionary mapping deme IDs to their coordinates, generated by `coords_to_deme_dict`.
    Optional inputs are dictionaries mapping ancestral population IDs to their constituent demes (`anc_demes_dict`), an dictionary of allele counts matrices for each deme (`ac_demes`), and dictionary of allele counts matrices for each ancestral population (`ac_anc`).


    Parameters
    ----------
    ac : np.ndarray
        An allele counts matrix for all individuals/demes.
    coords_dict : dict
        A dictionary mapping deme IDs to their coordinates, generated by [coords_to_deme_dict][utilities.coords_to_deme_dict].
    anc_demes_dict : dict, optional
        A dictionary mapping ancestral population IDs to their constituent demes. Defaults to None.
    ac_demes : dict, optional
        A dictionary mapping deme IDs to their allele counts matrices. Necessary if you want to calculate Fst or Dxy between demes. Defaults to None.
    ac_anc : dict, optional
        A dictionary mapping ancestral population IDs to their allel counts matrices. If provided, summary statistics are calculated within ancestral populations and not among them. Defaults to None.
    between_anc_pop_sumstats : bool, optional
        Whether to calculate Fst or Dxy between ancestral populations. Defaults to False.
    return_df : bool, optional
        Whether to return the summary statistics as a pandas DataFrame. Defaults to False.
    precision : int, optional
        The number of decimal places to round the summary statistics to. Defaults to 6.

    Returns
    -------
    dict
        A dictionary of summary statistics.



    Notes
    -----
    This function calculates the following summary statistics, either species-wide or per ancestral population, if provided:
    - Site Frequency Spectrum Hill numbers (q1 and q2), corrected for the number of sites
    - Pi (nucleotide diversity)
    - Tajima's D
    - Pairwise Dxy
        - If `between_anc_pop_sumstats` is True, also calculates pairwise Dxy and Hudson's FST between ancestral populations
    - Pairwise Hudson's FST
        - If `between_anc_pop_sumstats` is True, also calculates pairwise Dxy and Hudson's FST between ancestral populations
    - Isolation-by-distance slope and R2
    - Moran's I

    """

    # calculate scaled Hill numbers of the SFS, using Isaac Overcast's approach
    def _hill_number(freqs, order):
        if order == 0:
            return len(np.nonzero(freqs)[0])
        if order == 1:
            h1 = np.exp(entropy(freqs))
            return h1
        tot = float(np.sum(freqs))
        proportions = np.array(freqs[freqs > 0]) / tot
        prop_order = proportions**order
        h2 = np.sum(prop_order) ** (1 / (1 - order))
        return h2

    # get pairwise dxy for pops
    # deme IDs is a list of population IDs we want to calculate DXY for.
    def _calc_dxy(deme_ids):
        dxy = []
        ## also get the names of each pop pair to identify the raw dxy
        dxy_name = []

        # I'm using the coords_dict dictionary to iterate through the population names, so I can be sure that the DXY distances match up with the geographic distances
        for ac_ind in list(itertools.combinations(deme_ids, 2)):
            ac1 = ac_demes[ac_ind[0]]
            ac2 = ac_demes[ac_ind[1]]

            d = np.nanmean(allel.mean_pairwise_difference_between(ac1, ac2))
            d_name = f"dxy_{ac_ind[0]}_{ac_ind[1]}"

            dxy.append(d)
            dxy_name.append(d_name)
        return dxy, dxy_name

    # isolation-by-distance function
    ##### slope and r2 of gen ~ geo regression ##### (similar to mantel corr)
    def _calc_ibd(dxy, coords):

        # scale dxy according to Roussett 1997 (they used FST, but logic still follows)
        dxy = np.asarray(dxy)

        dxy_scaled = dxy / (1 - dxy)

        # get pairwise geographic distance for all pops

        # convert to radians
        long_rad = [radians(x[0]) for x in coords]
        lat_rad = [radians(y[1]) for y in coords]

        geometry = list(zip(long_rad, lat_rad))

        geometry = [list(_) for _ in geometry]

        dist = (
            haversine_distances(geometry) * 6371000 / 1000
        )  # multiply by Earth radius to get kilometers

        # get the upper diagonal of the distance matrix
        dist = dist[np.triu_indices(dist.shape[0], k=1)]

        # take the log10 of geographic distance to scale linearly for IBD regression
        logdist = np.log10(dist)

        dist_df = pd.DataFrame(data={"geo_dist": logdist, "dxy": dxy_scaled})

        dist_df = dist_df.dropna()

        # linear model, extracting the R2 and coefficient (slope)
        reg = linear_model.LinearRegression()

        # reshape X so it is 2D
        geo_dist = np.array(dist_df["geo_dist"])
        geo_dist = geo_dist.reshape(-1, 1)

        reg.fit(geo_dist, dist_df["dxy"])

        r2 = reg.score(geo_dist, dist_df["dxy"])
        b = reg.coef_[0]

        return b, r2

    # only calculate species-wide stats if there aren't ancestral populations
    if ac_anc is None:
        sfs = allel.sfs_folded(ac)

        # calculate the first 2 Hill numbers of the site frequency spectrum, scaling by the sample size
        sfs_h1 = _hill_number(sfs, 1) / len(sfs)
        sfs_h2 = _hill_number(sfs, 2) / len(sfs)

        # average pi across sites
        pi_raw = allel.mean_pairwise_difference(ac, fill=np.nan)
        pi = np.nanmean(pi_raw)
        # standard deviation of pi across sites
        pi_sd = np.nanstd(pi_raw)

        # tajima's D
        taj_d_raw = allel.moving_tajima_d(ac, size=100, step=10)
        taj_d = np.nanmean(taj_d_raw)
        taj_d_sd = np.nanstd(taj_d_raw)

        stat_dict = {
            "sfs_h1": sfs_h1,
            "sfs_h2": sfs_h2,
            "pi": pi,
            "pi_sd": pi_sd,
            "taj_d": taj_d,
            "taj_d_sd": taj_d_sd,
        }

        if ac_demes is not None:
            # calculate dxy
            dids = list(coords_dict.keys())

            dxy, dxy_name = _calc_dxy(dids)

            for name in range(len(dxy_name)):
                stat_dict[dxy_name[name]] = dxy[name]

    # otherwise, calculate stats per ancestral population
    else:
        stat_dict = {}

        # calc sfs Hill numbers, pi, and tajima's D
        for key in ac_anc:
            sfs = allel.sfs_folded(ac_anc[key])
            stat_dict[f"sfs_h1_a{key}"] = _hill_number(sfs, 1) / len(sfs)
            stat_dict[f"sfs_h2_a{key}"] = _hill_number(sfs, 2) / len(sfs)
            pi_raw = allel.mean_pairwise_difference(ac_anc[key])
            stat_dict[f"pi_a{key}"] = np.nanmean(pi_raw)
            stat_dict[f"pi_sd_a{key}"] = np.nanstd(pi_raw)
            taj_d_raw = allel.moving_tajima_d(ac_anc[key], size=100, step=10)
            stat_dict[f"taj_d_a{key}"] = np.nanmean(taj_d_raw)
            stat_dict[f"taj_d_sd_a{key}"] = np.nanstd(taj_d_raw)

        # Dxy and Hudson's FST for ancestral populations and Dxy for all pops.
        if between_anc_pop_sumstats:
            for ac_ind in list(itertools.combinations(list(ac_anc.keys()), 2)):
                ac1 = ac_anc[ac_ind[0]]
                ac2 = ac_anc[ac_ind[1]]
                d_raw = allel.mean_pairwise_difference_between(ac1, ac2)
                d = np.nanmean(d_raw)
                d_sd = np.nanstd(d_raw)
                stat_dict[f"dxy_a{ac_ind[0]}_{ac_ind[1]}"] = d
                stat_dict[f"dxy_sd_a{ac_ind[0]}_{ac_ind[1]}"] = d_sd
                fst, fst_se, _, _ = allel.average_hudson_fst(ac1, ac2, 100)
                stat_dict[f"fst_a{ac_ind[0]}_{ac_ind[1]}"] = fst
                stat_dict[f"fst_se_a{ac_ind[0]}_{ac_ind[1]}"] = fst_se

            if ac_demes is not None:
                # calculate dxy
                dids = list(coords_dict.keys())

                dxy, dxy_name = _calc_dxy(dids)

                for name in range(len(dxy_name)):
                    stat_dict[dxy_name[name]] = dxy[name]
            else:
                print(
                    "You need to provide allele counts matrices for demes to calculate Dxy between demes."
                )

    # calculate per-deme stats and spatial statistics if ac_demes is provided
    if ac_demes is not None:
        # get pi for all pops
        pi_demes = []
        for key in ac_demes:
            pi_demes.append(np.nanmean(allel.mean_pairwise_difference(ac_demes[key])))

        pi_demes_dict = dict(zip(ac_demes.keys(), pi_demes))

        # add pi to dictionary
        for key in pi_demes_dict:
            colname = f"pi_deme_{key}"
            stat_dict[colname] = pi_demes_dict[key]

            ####### Moran's I ########
        # convert the values of the coords_dict dictionary into a list
        coords = list(coords_dict.values())
        # convert to radians
        long_rad = [radians(x[0]) for x in coords]
        lat_rad = [radians(y[1]) for y in coords]
        ## geographic weights from Voronoi polygons
        coords_rad = np.array([long_rad, lat_rad]).transpose()

        weights = Voronoi(coords_rad, use_index=False)

        # array of per-pop pi, calculated earlier
        pi_array = np.array(pi_demes)
        moran_i = Moran(pi_array, weights)

        mi_i = moran_i.I

        stat_dict["morans_i"] = mi_i

        if anc_demes_dict is None:

            # convert the values of the coords_dict dictionary into a list
            coords = list(coords_dict.values())
            b, r2 = _calc_ibd(dxy, coords)
            stat_dict["ibd_slope"] = b
            stat_dict["ibd_r2"] = r2

        else:
            for key in anc_demes_dict:
                demes = anc_demes_dict[key]

                # get the coordinates associated with the demes
                coords_d = []
                for d in demes:
                    c = coords_dict[d]
                    coords_d.append(c)

                ## calculate dxy. This is super fast, so it's easier to do this than try weird subsetting to subset the full IBD matrix for the DXY values we want per admix pop
                dxy, dxy_name = _calc_dxy(demes)

                for name in range(len(dxy_name)):
                    stat_dict[dxy_name[name]] = dxy[name]

                ## calculate ibd
                b, r2 = _calc_ibd(dxy, coords_d)
                stat_dict[f"ibd_slope_a{key}"] = b
                stat_dict[f"ibd_r2_a{key}"] = r2

    # round all values
    stat_dict = {key: np.round(value, precision) for key, value in stat_dict.items()}

    if return_df:
        stat_df = pd.DataFrame(stat_dict, index=[0])
        return stat_df
    else:

        return stat_dict



""" Module for visualizing data."""

import numpy as np
import pandas as pd
from shapely.geometry import shape
import geopandas as gpd
import rasterio
import spaceprime
import contextily as ctx
from typing import List
from matplotlib import pyplot as plt


def validate_demographic_model(demo):
    if not hasattr(demo, "demes") or not hasattr(demo, "migration_array"):
        raise ValueError(
            "The demographic model must have 'demes' and 'migration_array' attributes. You may have forgotten to populate the spDemography object with `stepping_stone_2d()`."
        )


def validate_timestep(timestep, demo):
    if timestep < 0 or timestep >= len(demo.demes):
        raise ValueError(f"The timestep must be between 0 and {len(demo.demes) - 1}.")


def get_outgoing_migration_rates(
    demes: np.ndarray, mig_mat: np.ndarray, na_value: float = 1e-9
) -> pd.DataFrame:
    """
    Calculates the outgoing migration rates from each deme based on the migration matrix.

    Parameters
    ----------
    demes : np.ndarray
        An array representing the demes.
    mig_mat : np.ndarray
        A migration matrix representing the migration rates between demes.
    na_value : float, optional
        The value to use for missing migration rates. Defaults to 1e-9.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the outgoing migration rates from each deme.
    """
    # Get the shape of the demes array
    rows, cols = demes.shape
    # Create an empty list to store the migration data
    data = []

    # Iterate over each cell in the demes array
    for i in range(rows):
        for j in range(cols):
            # Calculate the current index based on the row and column
            current_index = i * cols + j
            # Get the deme size at the current cell
            deme_size = demes[i, j]

            # Skip cells with deme size below the specified threshold
            if deme_size <= na_value:
                continue

            # Calculate the migration rates for each direction
            mig_north = mig_mat[current_index, current_index - cols] if i > 0 else 0
            mig_east = mig_mat[current_index, current_index + 1] if j < cols - 1 else 0
            mig_south = (
                mig_mat[current_index, current_index + cols] if i < rows - 1 else 0
            )
            mig_west = mig_mat[current_index, current_index - 1] if j > 0 else 0

            # Append the migration data to the list
            data.append([deme_size, mig_north, mig_east, mig_south, mig_west])

    # Create a pandas DataFrame from the migration data
    return pd.DataFrame(
        data, columns=["deme_size", "mig_north", "mig_east", "mig_south", "mig_west"]
    )


def plot_model(
    demo: spaceprime.spDemography,
    raster: rasterio.DatasetReader,
    timestep: int,
    cmap: str = "viridis",
    legend: bool = True,
    tiles: str = "CartoDB positron",
):
    """
    Plots the demes and migration rates for a given timestep as an interactive map.

    Parameters
    ----------
    demo : spaceprime.spDemography
        The demographic model to plot.
    raster : rasterio.DatasetReader
        The raster dataset used to create the demes matrix(es).
    timestep : int
        The index of the desired timestep to plot.
    cmap : str, optional
        The colormap to use for the deme sizes. Defaults to 'viridis'.
    legend : bool, optional
        Whether to show the color legend. Defaults to True.
    tiles : str, optional
        The basemap tiles to use. Defaults to "CartoDB positron".

    Returns
    -------
    folium.folium.Map
        An interactive plot of the demes and migration rates.

    Example
    -------
    raster = rasterio.open("path/to/raster.tif")
    # Plot the model at timestep 1
    plot_model(demo, raster, 1)

    Notes
    -----
    Since this function returns a `folium` map object, you can further modify the map or save it to an HTML file with the `folium` library.
    """

    # check that the demography object contains demes and migration_array attributes
    validate_demographic_model(demo)

    # check if the number of timesteps is valid
    validate_timestep(timestep, demo)

    # Get the demes matrix for the specified timestep
    demes_matrix = demo.demes[timestep]

    # Mask the array to get only the valid data
    mask = demes_matrix > 1e-10

    # make sure each value is unique so rasterio doesn't merge cells
    array = np.arange(demes_matrix.size).reshape(demes_matrix.shape).astype(np.int32)

    shapes = rasterio.features.shapes(array, mask=mask, transform=raster.transform)

    # read the shapes as separate lists
    dummy_vals = []
    geometry = []
    for shapedict, value in shapes:
        dummy_vals.append(value)
        geometry.append(shape(shapedict))

    # build the gdf object over the two lists
    gdf = gpd.GeoDataFrame({"dummy": dummy_vals, "geometry": geometry}, crs=raster.crs)

    mig_df = get_outgoing_migration_rates(demes_matrix, demo.migration_array[timestep])

    # Add the migration rates to the GeoDataFrame
    gdf = pd.concat([gdf, mig_df], axis=1)

    # Add the deme index to the GeoDataFrame
    gdf["deme_index"] = gdf.index.to_list()

    # Plot the GeoDataFrames
    plot = gdf.explore(
        column="deme_size",
        tooltip=[
            "deme_size",
            "mig_north",
            "mig_east",
            "mig_south",
            "mig_west",
            "deme_index",
        ],
        cmap=cmap,
        legend=legend,
        tiles=tiles,
    )

    return plot


def plot_landscape(
    demo: spaceprime.spDemography,
    raster: rasterio.DatasetReader,
    timestep: int,
    cmap: str = "viridis",
    legend: bool = True,
    basemap: bool = False,
):
    """
    Plots a static map of a transformed landscape at the timestep of your choice.

    Parameters
    ----------
    demo : spaceprime.spDemography
        The demographic model to plot.
    raster : rasterio.DatasetReader
        The raster dataset used to create the demes matrix(es).
    timestep : int
        The timestep to plot.
    cmap : str, optional
        The colormap to use. Defaults to "viridis".
    legend : bool, optional
        Whether to show the colorbar legend. Defaults to True.
    basemap : bool, optional
        Whether to add a basemap. Requires an internet connection. Defaults to False.

    Returns
    -------
    matplotlib.axes.Axes
        A plot of the transformed landscape.

    Note
    ----
    Setting `basemap=True` requires an internet connection to download the basemap tiles. It may take some time to load the tiles depending on your internet speed.
    Since this function returns a `matplotlib` axes object, you can further modify the plot with the `matplotlib` library.
    """

    validate_demographic_model(demo)
    validate_timestep(timestep, demo)

    demes_matrix = demo.demes[timestep]
    mask = demes_matrix > 1e-10

    # plot the demes matrix
    array = np.arange(demes_matrix.size).reshape(demes_matrix.shape).astype(np.int32)

    shapes = rasterio.features.shapes(array, mask=mask, transform=raster.transform)

    dummy_vals = []
    geometry = []
    for shapedict, value in shapes:
        dummy_vals.append(value)
        geometry.append(shape(shapedict))

    gdf = gpd.GeoDataFrame({"dummy": dummy_vals, "geometry": geometry}, crs=raster.crs)
    # add the deme sizes to this gdf, making sure the masked values are omitted
    gdf["deme_size"] = demes_matrix[demes_matrix > 1e-10].flatten()

    if basemap:
        # use gdf.plot to plot the demes matrix
        plot = gdf.plot(column="deme_size", cmap=cmap, legend=legend, alpha=0.7)
        plot = ctx.add_basemap(plot, crs=gdf.crs, source=ctx.providers.CartoDB.Positron)
    else:
        plot = gdf.plot(column="deme_size", cmap=cmap, legend=legend)

    return plot


def plot_timeseries(demo: spaceprime.spDemography, times: List[float], units: str = ""):
    """
    Plots the total number of individuals across the landscape across time.

    Parameters
    ----------
    demo : spaceprime.spDemography
        The demographic model to plot.
    times : List[float]
        A list of times that each landscape timestep corresponds with. This can be in whatever unit you choose.
    units : str, optional
        The units of time the timesteps are specified in. Defaults to a blank string.

    Returns
    -------
    tuple
        A tuple containing the figure and axes objects of the plot.
    """

    validate_demographic_model(demo)

    # check if the times list is greater than 1
    if len(times) < 2:
        raise ValueError("The times list must contain more than one value.")

    # check if the length of the times list is equal to the number of timesteps
    if len(times) != len(demo.demes):
        raise ValueError(
            f"The length of the times list must be equal to the number of timesteps in the demographic model ({len(demo.demes)})."
        )

    # Get the total number of individuals at each timestep
    total_individuals = [np.nansum(d) for d in demo.demes]

    # Plot the timeseries as a line plot, using attractive formatting
    fig, ax = plt.subplots(figsize=(10, 6))
    ax.plot(times, total_individuals, marker="o", color="#224146")
    ax.set_xlabel(f"Time ({units})")
    ax.set_ylabel("Total number of individuals across the landscape")

    return fig, ax



"""
spaceprime is a python package to facilitate spatially explicit coalescent modeling in msprime.
Please see the documentation at https://connor-french.github.io/spaceprime
for more information.
"""

__author__ = """Connor French"""
__email__ = "french.connor.m@gmail.com"
__version__ = "0.0.5"


from spaceprime.utilities import (
    create_raster,
    raster_to_demes,
    calc_migration_matrix,
    split_landscape_by_pop,
    mtp_thresh_from_coords,
    coords_to_sample_dict,
    anc_to_deme_dict,
    coords_to_deme_dict,
)

from spaceprime.demography import (
    add_landscape_change,
    spDemography,
)

from spaceprime.simulation import (
    sim_ancestry,
    sim_mutations,
)

from spaceprime.plot import (
    plot_model,
    plot_landscape,
    plot_timeseries,
)

# optional install of analysis tools
try:
    from spaceprime.analysis import filter_gt, calc_sumstats
except ImportError:
    filter_gt = None
    calc_sumstats = None



"""
Module responsible for pre-processing data for creating demographic models
"""

from os import path
import numpy as np
import numpy.ma as ma
import geopandas as gpd
import pandas as pd
import rasterio
from rasterio.mask import mask
from scipy.interpolate import NearestNDInterpolator
from shapely.geometry import Point
from typing import Union, Optional, List, Tuple, Dict

import allel


## create_raster ##
def create_raster(
    data: np.ndarray,
    reference_raster: rasterio.DatasetReader,
    out_folder: str,
    out_prefix: str,
) -> None:
    """
    Creates a raster dataset from a numpy array and reference raster and writes it to a new GeoTiff file.
    The new raster dataset will have the same dimensions, crs, and transform as the reference raster.

    Parameters
    ----------
    data : np.ndarray
        The numpy array containing the data you want for the raster.
    reference_raster : rasterio.DatasetReader
        The reference rasterio DatasetReader object.
    out_folder : str
        The output folder location where the new raster dataset will be saved.
    out_prefix : str
        The prefix for the output file name.

    Returns
    -------
    None
        The function writes the new raster dataset to the output file location
    """
    with rasterio.open(
        path.join(out_folder, f"{out_prefix}.tif"),
        "w",
        driver="GTiff",
        height=data.shape[0],
        width=data.shape[1],
        count=(
            data.shape[2] if len(data.shape) == 3 else 1
        ),  # check if the data is 2D or 3D
        dtype=data.dtype,
        nodata=-9,
        crs=reference_raster.crs,
        transform=reference_raster.transform,
    ) as dataset:
        if len(data.shape) == 3:  # if data is 3D, write each layer to the dataset
            for i in range(data.shape[2]):
                dataset.write(data[:, :, i], i + 1)
        else:
            dataset.write(data, 1)


def raster_to_demes(
    raster: Union[np.ndarray, rasterio.DatasetReader],
    transformation: str = "linear",
    max_local_size: int = 1000,
    normalize: bool = False,
    threshold: Optional[float] = None,
    thresh_norm: bool = False,
    inflection_point: float = 0.5,
    slope: float = 0.05,
) -> np.ndarray:
    """
    Converts a raster to a 2D np.ndarray of deme sizes using either linear, threshold, or sigmoid transformation functions. For more detail about transformation functions, see [this brief overview](transformation-functions.qmd).
    Raster data should be continuous and positive.
    This function was created with the idea of taking in habitat suitability rasters scaled from 0 to 1, where 0 is no suitability and 1 is the highest suitability.
    However, it is flexible enough to accommodate other continuous rasters that can be coaxed to a 0 to 1 scale with the operation `(data - np.min(data)) / (np.max(data) - np.min(data))` by setting the `normalize` flag to `True`.


    Parameters
    ----------
    raster : Union[np.ndarray, rasterio.DatasetReader]
        The input raster data. It can be a numpy array or a rasterio DatasetReader with one or more layers.
    transformation : str, optional
        The transformation function to be used. Options are "linear", "threshold", and "sigmoid". Default is "linear".
    max_local_size : int, optional
        The maximum local deme size. Default is 1000.
    normalize : bool, optional
        Whether to normalize the raster data. Use if your data is not scaled from 0-1. Default is False.
    threshold : float, optional
        The threshold value for the "threshold" transformation method. Default is None.
    thresh_norm : bool, optional
        Whether to normalize the local deme size based on the average suitability above the threshold. This is useful when comparing thresholded simulations with linear or sigmoid simulations, to maintain similar landscape-wide population sizes across max_local_size values. Default is False.
    inflection_point : float, optional
        The inflection point for the "sigmoid" transformation method. Default is 0.5.
    slope : float, optional
        The slope value for the "sigmoid" transformation method. Default is 0.05.

    Returns
    -------
    np.ndarray
        An ndarray of deme sizes.
    """

    def check_raster_input(raster):
        if not isinstance(raster, (np.ndarray, rasterio.DatasetReader)):
            raise TypeError(
                "Invalid raster input. Expected numpy array or rasterio DatasetReader object."
            )

    def check_transformation_input(transformation):
        valid_transformations = ["linear", "threshold", "sigmoid"]
        if transformation not in valid_transformations:
            raise ValueError(
                f"Invalid transformation input. Expected one of {valid_transformations}."
            )

    check_raster_input(raster)
    check_transformation_input(transformation)

    # if the input is a rasterio object, read in the raster
    if isinstance(raster, rasterio.DatasetReader):
        d = raster.read(masked=True).filled(0)
    else:
        d = np.where(np.isnan(raster), 0, raster)
    if normalize:

        def _normalize(rast):
            return (rast - np.min(rast)) / (
                np.max(rast) - np.min(rast)
            )  # Normalize the raster data to a 0-1 scale

        d = _normalize(d)  # Apply normalization to the raster data

    if transformation == "linear":
        t = d * max_local_size  # Apply linear transformation to the raster data

        if threshold is not None:
            t[t < max_local_size * threshold] = 1e-10

        t[t > 1e-10] = np.ceil(
            t[t > 1e-10]
        )  # Round up the deme sizes to the nearest integer

    if transformation == "threshold":
        t = d  # Set the deme sizes to the raster data
        if thresh_norm:
            avg_sdm = np.nanmean(
                t[t >= threshold]
            )  # Calculate the average suitability above the threshold
            t[t >= threshold] = np.ceil(
                avg_sdm * max_local_size
            )  # Set the deme sizes above the threshold to the average multiplied by the maximum local size
        else:
            t[t >= threshold] = max_local_size

        t[t < threshold] = 1e-10

    if transformation == "sigmoid":

        def _sigmoid(x, a, b):
            y = 1.0 / (1.0 + np.ma.exp(-(x - a) / b))  # Sigmoid transformation function
            return y

        # vectorize the sigmoid function so it can be applied to the entire raster quickly
        sigmoid_v = np.vectorize(_sigmoid)
        t = (
            sigmoid_v(d, inflection_point, slope) * max_local_size
        )  # Apply sigmoid transformation to the raster data

        if threshold is not None:
            t[t < max_local_size * threshold] = 1e-10

        t[t > 1e-10] = np.ceil(
            t[t > 1e-10]
        )  # Round up the deme sizes to the nearest integer

    t = np.where(t < 1e-10, 1e-10, t)  # Set any deme sizes less than 1e-10 to 1e-10

    return t


######################
## TODO: expand this function take a rasterio DatasetReader object as input for the demes argument####
######################
def calc_migration_matrix(
    demes: np.ndarray, rate: float, scale: bool = True
) -> np.ndarray:
    """
    Calculates a migration matrix based on deme sizes and a global migration rate. The migration rate can be scaled based on population size or set as a constant.

    Parameters
    ----------
    demes : np.ndarray
        The 2D numpy array representing the deme sizes.
    rate : float
        The migration rate.
    scale : bool, optional
        Whether to scale the migration rate based on population size. Default is True.

    Returns
    -------
    np.ndarray
        The migration matrix as a 2D numpy array.
    """
    if not isinstance(demes, np.ndarray):
        raise TypeError("Invalid demes input. Expected numpy array.")
    if not isinstance(rate, float):
        raise TypeError("Invalid rate input. Expected float.")
    if not isinstance(scale, bool):
        raise TypeError("Invalid scale input. Expected bool.")

    d = demes.shape[0] * demes.shape[1]
    M = np.zeros((d, d))

    for i in range(demes.shape[0]):
        for j in range(demes.shape[1]):
            current_index = i * demes.shape[1] + j
            # check the neighboring demes and calculate the migration rates
            for di, dj in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                # check for edges
                if 0 <= i + di < demes.shape[0] and 0 <= j + dj < demes.shape[1]:
                    neighbor_index = (i + di) * demes.shape[1] + (j + dj)
                    if scale:
                        # calculate migration rate based on population size
                        # mig = donor / recipient * rate unless the pop size is zero
                        M[current_index, neighbor_index] = (
                            (demes[i + di, j + dj] / demes[i, j]) * rate
                            if (
                                (demes[i + di, j + dj] > 1e-9).all()
                                & (demes[i, j] > 1e-9).all()
                            )
                            else 0
                        )
                    else:
                        # use a constant migration rate
                        M[current_index, neighbor_index] = (
                            rate
                            if (
                                (demes[i + di, j + dj] > 1e-9).all()
                                & (demes[i, j] > 1e-9).all()
                            )
                            else 0
                        )

    return M


## split_landscape_by_pop ##
def split_landscape_by_pop(
    raster: rasterio.DatasetReader,
    coordinates: Union[List[Tuple[float, float]], gpd.GeoDataFrame],
    anc_pop_id: List[Union[int, np.integer]],
    band_index: int = 1,
    mask_rast: bool = False,
) -> np.ma.MaskedArray:
    """
    Uses nearest-neighbor interpolation to classify a landscape raster based on the ancestral population assigned to sampled individuals.
    This function takes in a raster and a list of coordinates and ancestral population IDs assigned to each individual in the empirical data set.
    It then interpolates the population IDs across the landscape and returns the new raster as a masked array.

    Parameters
    ----------
    raster : rasterio.DatasetReader
        The rasterio DatasetReader object representing the landscape raster that you want to divide.
    coordinates : Union[List[Tuple[float, float]], gpd.GeoDataFrame]
        A list of (x, y) coordinates or a geopandas GeoDataFrame representing the coordinates assigned to each individual in the empirical data set.
    anc_pop_id : List[Union[int, np.integer]]
        A list of ancestral population IDs assigned to each empirical individual[^1].
    band_index : int, optional
        The index of the raster to read in. Default is 1. Note- rasterio begins indexing at 1 for raster bands.
    mask_rast : bool, optional
        Whether to mask the interpolation by the landscape. Default is False.

    Returns
    -------
    np.ma.MaskedArray
        The new population assignment raster as a masked array.

    Notes
    -----
    [^1]: These IDs are assigned to each empirical individual typically based on genetic clustering methods like STRUCTURE or PCA. The IDs are used to assign individuals to ancestral populations in the landscape.
    """
    # check if coordinates is a list of tuples or a geopandas GeoDataFrame
    if not isinstance(coordinates, list) and not isinstance(
        coordinates, gpd.GeoDataFrame
    ):
        raise TypeError(
            "The coordinates must be a list of tuples or a geopandas GeoDataFrame."
        )

    if isinstance(coordinates, gpd.GeoDataFrame):
        # convert the geometry column to a list of tuples
        coordinates = list(coordinates.geometry.apply(lambda geom: (geom.x, geom.y)))

    if isinstance(anc_pop_id, pd.Series):
        raise TypeError(
            "The anc_pop_id must be a list of whole numbers. If using a pandas Series, convert it to a list using the to_list() method."
        )

    if not isinstance(anc_pop_id, list) or not all(
        isinstance(x, (int, np.integer)) for x in anc_pop_id
    ):
        raise TypeError("The anc_pop_id must be a list of whole numbers.")

    # read in the raster. This imports the raster as a numpy array
    r2 = raster.read(band_index, masked=True)

    # get the x,y indices of the empirical sampled cells
    indices_x = []
    indices_y = []

    for xy in coordinates:

        # mask requires an iterable as input, so I just repeated the two Point geometries in a list. Mask returns a single value since they overlap in the same place
        pt2 = [Point(xy), Point(xy)]

        # mask the raster with the points
        out_image = mask(raster, pt2, nodata="nan", filled=False)

        # oi first raster
        oi = out_image[0][0]

        # get the locality index
        indices_tup = ma.nonzero(oi)

        indices_x.append(indices_tup[0])
        indices_y.append(indices_tup[1])

    indices_x = np.concatenate(indices_x)
    indices_y = np.concatenate(indices_y)

    # check that the number of indices matches the number of population IDs. If they don't match, this means that some coordinates did not overlap with the raster
    if len(indices_x) != len(anc_pop_id):
        raise ValueError(
            "Some coordinates did not overlap with the raster. Make sure all sampling coordinates overlap with a cell in the raster."
        )

    # get all of the x, y indices of the input raster
    r_x, r_y = np.indices(raster.shape)

    interp = NearestNDInterpolator(list(zip(indices_x, indices_y)), anc_pop_id)
    z = interp(r_x, r_y)
    z = z.astype(int)

    # apply mask of the SDM landscape
    if mask_rast:
        z = ma.masked_array(z, r2.mask, fill_value=-9, dtype=int)

    return z  # Return the new raster as a masked array


def mtp_thresh_from_coords(
    raster: rasterio.DatasetReader,
    coordinates: Union[List[Tuple[float, float]], gpd.GeoDataFrame],
) -> float:
    """
    This function takes the coordinates of empirical sampling localities, finds which raster cells they belong to, extracts the values of the first layer for those localities, and finds the minimum value.
    This value is the maximum threshold value to determine a presence vs absence in a threshold transformation.
    If the threshold is set any higher, empirical sampling localities will not be sampled in the simulations.

    Parameters
    ----------
    raster : rasterio.DatasetReader
        The rasterio DatasetReader object representing the raster data containing the suitability values.
    coordinates : Union[List[Tuple[float, float]], gpd.GeoDataFrame]
        The longitude, latitude coordinates of the empirical sampling localities as a list of coordinate pair tuples or a geopandas GeoDataFrame.

    Returns
    -------
    float
        The maximum threshold value to determine a presence vs absence in a threshold transformation.

    Raises
    ------
    TypeError
        If the coordinates input is not a list, geopandas GeoDataFrame, or pandas DataFrame.
    """
    if isinstance(coordinates, list):
        xy = [Point(xy) for xy in coordinates]
    elif isinstance(coordinates, gpd.GeoDataFrame):
        xy = coordinates.geometry
    else:
        raise TypeError(
            "Invalid coordinates input. Expected list of coordinate pairs or geopandas GeoDataFrame."
        )

    # Mask the first layer with the coordinates
    out_image = mask(raster, xy)
    # Find the minimum value of the masked first layer
    max_thresh = np.nanmin(out_image[0][0])

    return max_thresh


def coords_to_sample_dict(
    raster: Union[np.ndarray, rasterio.DatasetReader],
    coordinates: Union[List[Tuple[float, float]], gpd.GeoDataFrame],
    individual_ids: Optional[List[str]] = None,
    vcf_path: Optional[str] = None,
) -> Tuple[Dict[int, int], Dict[int, np.ndarray], Optional[Dict[int, np.ndarray]]]:
    """
    Convert sample coordinates to sample dictionaries for simulation and analysis. Can optionally include empirical data, which is accepted as a path to a VCF file.

    This function takes a raster, a list of coordinates, and optional individual IDs and VCF path.
    It masks the raster with the given coordinates, retrieves the cell IDs for each individual's locality,
    and returns two dictionaries: a sample dictionary containing the number of individuals to sample from the simulation, and a sample dictionary containing the range of individual indices for each cell ID.
    The first dictionary is used to sample individuals from the simulation, and the second dictionary is used to calculate genetic summary statistics from the sampled individuals.

    Parameters
    ----------
    raster : Union[np.ndarray, rasterio.DatasetReader]
        The raster data as a numpy array or rasterio DatasetReader object.
    coordinates : Union[List[Tuple[float, float]], gpd.GeoDataFrame]
        A list of (x, y) coordinates or a geopandas GeoDataFrame.
    individual_ids : Optional[List[str]], optional
        A list of individual IDs corresponding to those in the VCF file, by default None.
    vcf_path : Optional[str], optional
        The path to the VCF file, by default None.

    Returns
    -------
    Tuple[Dict[int, int], Dict[int, np.ndarray], Optional[Dict[int, np.ndarray]]]
        A tuple containing two or three dictionaries.
        The first dictionary contains the number of individuals to sample from the simulation for each cell ID.
        The second dictionary contains the indices of individuals for each cell ID.
        The third, optional dictionary contains the indices of individuals in the VCF file for each cell ID.

    """
    # check for individual ids and vcf path
    if individual_ids is None:
        try:
            if vcf_path is not None:
                raise ValueError
        except ValueError:
            print(
                "When a VCF path is provided, individual IDs corresponding to those in the VCF are expected. Please provide those IDs."
            )
    elif vcf_path is None:
        try:
            if individual_ids is not None:
                raise ValueError
        except ValueError:
            print(
                "When individual IDs are provided, a VCF file is expected. Please provide a path to your VCF file."
            )
    else:
        deme_dict_empirical = None

    # get the cell that each individual belongs to
    ## I have to iterate through each locality to get the cell ID for each individual locality. Masking everything returns the cell IDs out of order.
    cell_list = []

    # check if coordinates is a geopandas dataframe
    if isinstance(coordinates, gpd.GeoDataFrame):
        # convert the geometry column to a list of tuples
        coordinates = list(coordinates.geometry.apply(lambda geom: (geom.x, geom.y)))

    for xy in coordinates:
        # mask requires an iterable as input, so I just repeated the two Point geometries in a list. Mask returns a single value since they overlap in the same place
        pt2 = [Point(xy), Point(xy)]

        # mask the raster with the points
        out_image = mask(raster, pt2, nodata="nan", filled=False)

        # turn into 1D array
        oi_1d = out_image[0][0].ravel()

        # # get the indices of the locality and append
        cell_id = ma.nonzero(oi_1d)[0]
        cell_list.append(cell_id)

    cell_id_array = np.concatenate(cell_list, axis=0)

    # get the number of individuals to sample from the simulation
    deme_dict_sim = {}
    for cid in np.unique(cell_id_array):
        num_inds = np.sum(cell_id_array == cid)
        deme_dict_sim[int(cid)] = num_inds

    # get the range of indices for each cell id
    deme_dict_sim_long = {}
    unique_ids = np.unique(cell_id_array)
    for i in range(len(unique_ids)):
        cid = unique_ids[i]
        if i > 0:
            cid_prev = unique_ids[i - 1]
            last_ind_prev = deme_dict_sim_long[cid_prev][
                len(deme_dict_sim_long[cid_prev]) - 1
            ]  # get the last index of the previous cell id
            deme_dict_sim_long[int(cid)] = np.array(
                range(last_ind_prev + 1, last_ind_prev + deme_dict_sim[cid] + 1)
            )  # get the range of indices for the current cell id
        else:
            deme_dict_sim_long[int(cid)] = np.array(range(deme_dict_sim[cid]))

    if individual_ids is not None and vcf_path is not None:
        # read in the vcf
        callset = allel.read_vcf(vcf_path)

        # get the indices of each individual in the vcf
        ind_indices = []
        for cid in individual_ids:
            index = list(callset["samples"]).index(cid)
            ind_indices.append(index)

        ind_indices_array = np.array(ind_indices)

        # make each cell id a key and the individuals that belong to that cell id the values
        deme_dict_empirical = {}
        for cid in np.unique(cell_id_array):
            id_inds = np.where(cell_id_array == cid)
            id_ind_indices = ind_indices_array[id_inds]
            deme_dict_empirical[int(cid)] = id_ind_indices
    else:
        deme_dict_empirical = None

    return deme_dict_sim, deme_dict_sim_long, deme_dict_empirical


def anc_to_deme_dict(
    anc_pops: np.ndarray, deme_dict: Dict[int, int]
) -> Dict[int, List[int]]:
    """
    Converts the ancestral population assignments of demes into a dictionary mapping ancestral population IDs to deme indices.

    Parameters
    ----------
    anc_pops : np.ndarray
        An array containing the ancestral population assignments of all demes across the landscape.
    deme_dict : Dict[int, int]
        A dictionary mapping deme indices to the number of individuals being sampled from each deme.

    Returns
    -------
    Dict[int, List[int]]
        A dictionary mapping each ancestral population ID to the range of assigned deme indices.
    """
    # unravel the raster into one dimension for indexing
    r_1d = anc_pops.ravel()

    ap_dict = {}
    for key in deme_dict:
        k = r_1d[key]
        if k in ap_dict.keys():
            ap_dict[k].append(key)
        else:
            ap_dict[k] = [key]

    return ap_dict


def coords_to_deme_dict(
    raster: rasterio.DatasetReader,
    coordinates: Union[List[Tuple[float, float]], gpd.GeoDataFrame],
) -> Dict[int, List[float]]:
    """
    Finds the cells a given set of coordinates belong to in a raster and returns a dictionary mapping the cell indices to the centroid coordinates of those cells.
    Because the cells correspond with demes in the 2D stepping stone models, the cell indices are considered deme indices.
    The coordinates typically correspond to empirical data that the simulations need to be sampled from.

    Parameters
    ----------
    raster : rasterio.DatasetReader
        The raster data as a rasterio DatasetReader object.
    coordinates : Union[List[Tuple[float, float]], gpd.GeoDataFrame]
        A list of (x, y) coordinates or a geopandas GeoDataFrame.

    Returns
    -------
    Dict[int, List[float]]
        A dictionary mapping deme indices to their corresponding coordinates.
    """

    # check if raster is a rasterio.DatasetReader object
    if not isinstance(raster, rasterio.DatasetReader):
        raise TypeError("The raster must be a rasterio.DatasetReader object.")

    # check if coordinates is a list of tuples or a geopandas GeoDataFrame
    if not isinstance(coordinates, list) and not isinstance(
        coordinates, gpd.GeoDataFrame
    ):
        raise TypeError(
            "The coordinates must be a list of tuples or a geopandas GeoDataFrame."
        )

    # check if coordinates is a geopandas dataframe
    if isinstance(coordinates, gpd.GeoDataFrame):
        # convert the geometry column to a list of tuples
        coordinates = list(coordinates.geometry.apply(lambda geom: (geom.x, geom.y)))

    # get the population IDs
    cell_dict = {}
    for xy in coordinates:
        # mask requires an iterable as input, so I just repeated the two Point geometries in a list. Mask returns a single value since they overlap in the same place
        pt2 = [Point(xy), Point(xy)]

        # mask the raster with the points
        out_image = mask(raster, pt2, nodata="nan", filled=False)

        # GET CELL INDICES
        # turn into 1D array
        oi_1d = out_image[0][0].ravel()

        # get the indices of the locality and append
        cell_id = ma.nonzero(oi_1d)[0][0]

        # GET CELL COORDINATES
        # get the indices of the locality
        cell_ind = [tuple(_) for _ in np.transpose(ma.nonzero(out_image[0][0]))]

        rows, cols = zip(
            *cell_ind
        )  # unzip the cell indices to separate row and col lists
        coords_tuple = rasterio.transform.xy(raster.transform, rows, cols)

        # I have to convert the x and y coords to a list instead of a tuple, because the transform.xy function returns a tuple where each long and lat is a single-value list, which makes indexing later convoluted.
        x = coords_tuple[0][0]
        y = coords_tuple[1][0]

        coords = [x, y]

        cell_dict[cell_id] = coords

    return cell_dict



"""Console script for spaceprime."""

import argparse
import os
import yaml
import rasterio
import msprime
import pandas as pd
import numpy as np
from numpy.random import default_rng
from geopandas import GeoDataFrame
from shapely.geometry import Point
import time
from multiprocessing import Pool

from . import utilities
from . import demography
from . import analysis

# set up logging
import logging

logging.basicConfig(filename="iddc.log", level=logging.INFO)
# write logs to a file


logging.info("spaceprime CLI script started")


# Check if list arguments have more than two elements
def check_argument_length(arg, max_length):
    if arg is not None and len(arg) > max_length:
        raise ValueError(f"{arg} argument can have a maximum of {max_length} elements")


# Check if list arguments are lists, else convert to lists
def check_list_argument(arg):
    if arg is None:
        return None
    elif isinstance(arg, list):
        return arg
    else:
        return [arg]


# Assuming your columns are named 'longitude' and 'latitude'
def read_coords(file_path):
    # check if coords file exists
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"Coordinates file {file_path} not found")
    # check if coords file is a CSV
    if not file_path.endswith(".csv"):
        raise ValueError(f"Coordinates file {file_path} must be a CSV")
    df = pd.read_csv(file_path)
    # Check if 'longitude' and 'latitude' columns exist
    if "longitude" not in df.columns or "latitude" not in df.columns:
        raise ValueError("DataFrame must contain 'longitude' and 'latitude' columns.")
    gdf = GeoDataFrame(
        df, geometry=[Point(xy) for xy in zip(df.longitude, df.latitude)]
    )
    return gdf


# read in individual IDs if a path is provided, otherwise use the list
# also perform a series of checks
def read_individuals(args, coords):
    if args.individuals is not None:
        if isinstance(args.individuals, str):
            if not os.path.exists(args.individuals):
                raise FileNotFoundError(
                    f"Individuals file {args.individuals} not found"
                )
            if not args.individuals.endswith(".csv"):
                raise ValueError(f"Individuals file {args.individuals} must be a CSV")
            individuals = pd.read_csv(args.individuals)
            if "individual_id" not in individuals.columns:
                raise ValueError(
                    "Individuals file must contain a column named 'individual_id'"
                )
            individuals = individuals["individual_id"].tolist()
        else:
            individuals = args.individuals

        if len(individuals) != len(coords):
            raise ValueError(
                "Number of individuals must be the same as the number of coordinates"
            )

    return individuals


def read_anc_pop_id(anc_pop_id):
    if anc_pop_id is not None and not len(anc_pop_id) > 1:
        if not os.path.exists(anc_pop_id[0]):
            raise FileNotFoundError(f"anc_pop_id file {anc_pop_id[0]} not found")
        else:
            # Read the CSV file using pandas
            df = pd.read_csv(anc_pop_id[0])
            # Check if 'anc_pop_id' column exists in the CSV file
            if "anc_pop_id" not in df.columns:
                raise ValueError("CSV file does not contain 'anc_pop_id' column")
            # Return the 'anc_pop_id' column as a list
            return df["anc_pop_id"].tolist()
    else:
        return anc_pop_id


# for outputting a map of genetic diversity
def get_map_dict(m, min_num_inds=2, sample_num=2):
    nzp = []

    for i in range(len(m.populations)):
        if (
            m.populations[i].initial_size >= min_num_inds
            and m.populations[i].name != "ANC"
        ):
            nzp.append(m.populations[i].name)

    map_dict = {key: sample_num for key in nzp}
    return map_dict


# get coalescent times for each deme if the map is True
def get_coal_times(tseq, raster, num_anc_pops, sample_num=2, ploidy=2):
    # account for merged ancestral population
    if num_anc_pops > 1:
        num_anc_pops += 1

    coal_list = []
    for j in range(tseq.num_populations):
        # Get the samples corresponding to this population
        samples = tseq.samples(population=j)
        if len(samples) == ploidy * sample_num:
            # Simplify the tree sequence to just these samples
            ts_pop = tseq.simplify(samples=samples)
            div_pi = ts_pop.diversity()
            coal_list.append(div_pi)
        else:
            coal_list.append(-1)

    coal_1d = np.array(coal_list)[:-num_anc_pops]

    coal_array = np.reshape(coal_1d, newshape=raster.shape)

    return coal_array


def setup_demography(
    raster,
    coords,
    max_local_size,
    threshold,
    inflection_point,
    slope,
    mig_rate,
    scale,
    anc_pop_id,
    timesteps,
    anc_sizes,
    merge_time,
    anc_merge_time,
    anc_merge_size,
    anc_mig_rate,
):
    # convert raster to demes
    demes = utilities.raster_to_demes(
        raster=raster,
        max_local_size=max_local_size,
        threshold=threshold,
        inflection_point=inflection_point,
        slope=slope,
    )
    # split landscape by population
    if anc_pop_id is not None:
        anc_pop_mat = utilities.split_landscape_by_pop(
            raster=raster, coordinates=coords, anc_pop_id=anc_pop_id
        )
    else:
        anc_pop_mat = None

    # create demography
    d = demography.stepping_stone_2d(
        d=demes, rate=mig_rate, scale=scale, timesteps=timesteps
    )

    # add ancestral populations
    d = demography.add_ancestral_populations(
        model=d,
        anc_sizes=anc_sizes,
        merge_time=merge_time,
        anc_id=anc_pop_mat,
        anc_merge_times=anc_merge_time,
        anc_merge_sizes=anc_merge_size,
        migration_rate=anc_mig_rate,
    )

    return d


def get_random_value(arg):
    rng = default_rng()
    if arg is None:
        return None
    elif isinstance(arg, (int, float)):
        return arg
    elif len(arg) == 1:
        return arg[0]
    elif isinstance(arg[0], list):
        if isinstance(arg[0][0], int):
            return [rng.integers(a, b + 1) for a, b in arg]
        else:
            return [rng.uniform(a, b) for a, b in arg]
    else:
        if isinstance(arg[0], int):
            return rng.integers(arg[0], arg[1] + 1)
        else:
            return rng.uniform(arg[0], arg[1])


def generate_param_combinations(args):
    param_combos = []
    for _ in range(args.num_param_combos):
        combo = {}
        # max_local_size
        combo["max_local_size"] = get_random_value(args.max_local_size)
        # threshold
        combo["threshold"] = get_random_value(args.threshold)
        # inflection_point
        combo["inflection_point"] = get_random_value(args.inflection_point)
        # slope
        combo["slope"] = get_random_value(args.slope)
        # mig_rate
        combo["mig_rate"] = get_random_value(args.mig_rate)
        # anc_sizes
        if args.anc_sizes is not None:
            combo["anc_sizes"] = get_random_value(args.anc_sizes)
        else:
            combo["anc_sizes"] = None
        # merge_time
        combo["merge_time"] = get_random_value(args.merge_time)
        # anc_merge_time
        if args.anc_merge_time is not None:
            combo["anc_merge_time"] = get_random_value(args.anc_merge_time)
        else:
            combo["anc_merge_time"] = None
        # anc_merge_size
        if args.anc_merge_size is not None:
            combo["anc_merge_size"] = get_random_value(args.anc_merge_size)
        else:
            combo["anc_merge_size"] = None
        # anc_mig_rate
        if args.anc_mig_rate is not None:
            combo["anc_mig_rate"] = get_random_value(args.anc_mig_rate)
        else:
            combo["anc_mig_rate"] = None
        # mutation_rate
        combo["mutation_rate"] = get_random_value(args.mutation_rate)
        # recombination_rate
        combo["recombination_rate"] = get_random_value(args.recombination_rate)

        param_combos.append(combo)

    return param_combos


def run_simulation(combo, args):
    rng = default_rng()
    logging.info(f"Running simulation with parameters: {combo}")
    # read in raster
    r = rasterio.open(args.raster)

    # read in the coordinates
    coords = read_coords(args.coords)

    # read in individuals
    individuals = read_individuals(args, coords)

    # read in anc_pop_id
    anc_pop_id = read_anc_pop_id(args.anc_pop_id)

    # sample dictionaries for ancestry sims and optionally for sumstats
    sample_dicts = utilities.coords_to_sample_dict(r, coords)

    demo_id = f"demo_{rng.integers(0, 2**30)}"
    logging.info(f"Setting up demography with ID: {demo_id}")
    d = setup_demography(
        raster=r,
        coords=coords,
        max_local_size=combo["max_local_size"],
        threshold=combo["threshold"],
        inflection_point=combo["inflection_point"],
        slope=combo["slope"],
        mig_rate=combo["mig_rate"],
        scale=args.scale,
        anc_pop_id=anc_pop_id,
        timesteps=args.timesteps,
        anc_sizes=combo["anc_sizes"],
        merge_time=combo["merge_time"],
        anc_merge_time=combo["anc_merge_time"],
        anc_merge_size=combo["anc_merge_size"],
        anc_mig_rate=combo["anc_mig_rate"],
    )

    print("Finished setting up demography")
    logging.info("Beginning tree sequence simulations")
    start_time = time.time()
    print("Beginning tree sequence simulations")

    # if map is True, return a dictionary mapping samples to all nonzero demes
    # replace the sample dictionary with this dictionary
    if args.map:
        if args.threshold is not None:
            min_num_inds = np.floor(args.threshold * combo["max_local_size"]).astype(
                int
            )
        else:
            min_num_inds = 2

        samples = get_map_dict(d, min_num_inds=min_num_inds)
    else:
        samples = sample_dicts[0]

    for ncoal in range(args.num_coalescent_sims):
        rng = default_rng()
        # set a new random seed for each ancestry simulation
        ancestry_seed = rng.integers(0, 2**30)
        mutation_seed = rng.integers(0, 2**30)

        # simulate tree sequence
        ts = msprime.sim_ancestry(
            samples=samples,
            demography=d,
            sequence_length=args.seq_length,
            recombination_rate=combo["recombination_rate"],
            ploidy=args.ploidy,
            random_seed=ancestry_seed,
            record_provenance=False,
        )

        # simulate mutations
        ts = msprime.sim_mutations(
            ts, rate=combo["mutation_rate"], random_seed=mutation_seed
        )

        end_time = time.time()
        elapsed_time = end_time - start_time

        # write metadata to csv
        metadata = {
            "ancestry_seed": ancestry_seed,
            "mutation_seed": mutation_seed,
            "max_local_size": (
                str(combo["max_local_size"])
                if isinstance(combo["max_local_size"], list)
                else combo["max_local_size"]
            ),
            "threshold": (
                str(combo["threshold"])
                if isinstance(combo["threshold"], list)
                else combo["threshold"]
            ),
            "inflection_point": (
                str(combo["inflection_point"])
                if isinstance(combo["inflection_point"], list)
                else combo["inflection_point"]
            ),
            "slope": (
                str(combo["slope"])
                if isinstance(combo["slope"], list)
                else combo["slope"]
            ),
            "mig_rate": (
                str(combo["mig_rate"])
                if isinstance(combo["mig_rate"], list)
                else combo["mig_rate"]
            ),
            "anc_sizes": (
                str(combo["anc_sizes"])
                if isinstance(combo["anc_sizes"], list)
                else combo["anc_sizes"]
            ),
            "merge_time": (
                str(combo["merge_time"])
                if isinstance(combo["merge_time"], list)
                else combo["merge_time"]
            ),
            "anc_merge_time": (
                str(combo["anc_merge_time"])
                if isinstance(combo["anc_merge_time"], list)
                else combo["anc_merge_time"]
            ),
            "anc_merge_size": (
                str(combo["anc_merge_size"])
                if isinstance(combo["anc_merge_size"], list)
                else combo["anc_merge_size"]
            ),
            "anc_mig_rate": (
                str(combo["anc_mig_rate"])
                if isinstance(combo["anc_mig_rate"], list)
                else combo["anc_mig_rate"]
            ),
            "mutation_rate": (
                str(combo["mutation_rate"])
                if isinstance(combo["mutation_rate"], list)
                else combo["mutation_rate"]
            ),
            "recombination_rate": (
                str(combo["recombination_rate"])
                if isinstance(combo["recombination_rate"], list)
                else combo["recombination_rate"]
            ),
            "demo_id": demo_id,
            "coal_rep": ncoal,
            "sim_time": elapsed_time,
        }
        metadata_file = os.path.join(args.out_folder, f"{args.out_prefix}_metadata.csv")
        if os.path.exists(metadata_file):
            pd.DataFrame(metadata, index=[0]).to_csv(
                metadata_file, mode="a", header=False, index=False
            )
        else:
            pd.DataFrame(metadata, index=[0]).to_csv(metadata_file, index=False)

        logging.info("Beginning coalescent array calculations")
        if args.map:
            if anc_pop_id is not None:
                coal_array = get_coal_times(
                    ts,
                    r,
                    len(set(anc_pop_id)),
                    ploidy=args.ploidy,
                    sample_num=2,
                )
            else:
                coal_array = get_coal_times(ts, r, 1, ploidy=args.ploidy, sample_num=2)

            utilities.create_raster(
                coal_array,
                r,
                out_folder=args.out_folder,
                out_prefix=f"{args.out_prefix}_diversity_map_{ancestry_seed}",
            )
        logging.info("Finished coalescent array calculations and wrote to file")

        # only output other files if args.map is False
        if not args.map:
            # if out_type is 0 or 3, write tree sequence to file
            if args.out_type == 0 or args.out_type == 3:
                ts_file = os.path.join(
                    args.out_folder, f"{args.out_prefix}_ancestry_{ancestry_seed}.trees"
                )
                ts.dump(ts_file)

            # if out_type is 1 or 3, write VCF to file
            if args.out_type == 1 or args.out_type == 3:
                vcf_file = os.path.join(
                    args.out_folder, f"{args.out_prefix}_vcf_{ancestry_seed}.vcf"
                )
                ts.write_vcf(vcf_file, ploidy=args.ploidy, individual_names=individuals)

            # if out_type is 2 or 3, calculate summary statistics
            #### NEED TO FINISH THIS ####
            if args.out_type == 2 or args.out_type == 3:
                # convert tree sequence to a genotype matrix
                gt = ts.genotype_matrix()
                # get necessary dictionaries
                coords_dict = utilities.coords_to_deme_dict(r, coords)

                # split landscape by population if within_anc_pop_sumstats is True or if between_anc_pop_sumstats is True
                if args.within_anc_pop_sumstats or args.between_anc_pop_sumstats:
                    anc_pop_mat = utilities.split_landscape_by_pop(
                        r, coords, anc_pop_id
                    )
                    deme_dict_anc = utilities.anc_to_deme_dict(
                        anc_pop_mat, sample_dicts[1]
                    )
                else:
                    deme_dict_anc = None

                # filter genotype matrix
                _, ac_filt, ac_demes, ac_anc = analysis.filter_gt(
                    gt,
                    deme_dict_inds=sample_dicts[1],
                    deme_dict_anc=deme_dict_anc,
                    missing_data_perc=args.missing_data_perc,
                    r2_thresh=args.r2_thresh,
                    filter_monomorphic=args.filter_monomorphic,
                    filter_singletons=args.filter_singletons,
                )

                # calculate summary statistics
                sumstats = analysis.calc_sumstats(
                    ac=ac_filt,
                    coords_dict=coords_dict,
                    anc_demes_dict=deme_dict_anc,
                    ac_demes=ac_demes,
                    ac_anc=ac_anc,
                    between_anc_pop_sumstats=args.between_anc_pop_sumstats,
                    return_df=True,
                    precision=6,
                )

                # add columns containing the ancestry and mutation seeds
                sumstats["ancestry_seed"] = ancestry_seed
                sumstats["mutation_seed"] = mutation_seed

                # write summary statistics to file
                sumstats_file = os.path.join(
                    args.out_folder, f"{args.out_prefix}_sumstats.csv"
                )
                if os.path.exists(sumstats_file):
                    sumstats.to_csv(sumstats_file, mode="a", header=False, index=False)
                else:
                    sumstats.to_csv(sumstats_file, index=False)

        print("Finished simulating tree sequences")


def main():
    """Console script for spaceprime."""
    parser = argparse.ArgumentParser()
    # global arguments
    parser.add_argument(
        "-p",
        "--params",
        type=str,
        default=None,
        help="Path to YAML file containing parameters. If provided, command line arguments will be ignored. Download a template from the spaceprime documentation. Default is None.",
    )
    parser.add_argument(
        "-r",
        "--raster",
        type=str,
        help="Path to raster file. Can be any raster format that rasterio can read.",
    )
    parser.add_argument(
        "-co",
        "--coords",
        type=str,
        help="Path to file containing sampling coordinates. Must be a CSV file containing columns 'longitude' and 'latitude'.",
    )
    parser.add_argument(
        "-i",
        "--individuals",
        type=lambda x: x.split(",") if "," in x else x,
        default=None,
        help="Either a list or a path to file containing individual IDs. If a CSV, must contain a column named 'individual_id'. Default is None.",
    )
    # Demography Setup
    demography_parser = parser.add_argument_group("Demography Setup")
    demography_parser.add_argument(
        "-n",
        "--normalize",
        type=bool,
        default=False,
        help="Normalize the raster to [0, 1] range. Default is False.",
    )
    demography_parser.add_argument(
        "-t",
        "--transformation",
        type=str,
        default="linear",
        help="Transformation to apply to raster. Options are 'linear', 'threshold', 'sigmoid'. Default is 'linear'.",
    )
    demography_parser.add_argument(
        "-mls",
        "--max_local_size",
        nargs="+",
        type=int,
        default=[1000],
        help="Maximum size of local demes. Accepts a single int or a pair of ints. Default is [1000].",
    )
    demography_parser.add_argument(
        "-th",
        "--threshold",
        nargs="+",
        type=float,
        default=None,
        help="Threshold value for a thresholded transformation. Accepts a single float or a pair of floats. Default is None.",
    )
    demography_parser.add_argument(
        "-ip",
        "--inflection_point",
        nargs="+",
        type=float,
        default=[0.5],
        help="Inflection point value for a sigmoid transformation. Accepts a single int or a pair of ints. Default is [0.5].",
    )
    demography_parser.add_argument(
        "-s",
        "--slope",
        nargs="+",
        type=float,
        default=[0.05],
        help="Slope value for a sigmoid transformation. Accepts a single int or a pair of ints. Default is [0.05].",
    )
    demography_parser.add_argument(
        "-m",
        "--mig_rate",
        nargs="+",
        type=float,
        default=None,
        help="Migration rate between demes. Accepts a single int or a pair of ints. Default is None.",
    )
    demography_parser.add_argument(
        "-sc",
        "--scale",
        type=bool,
        default=True,
        help="Scale the migration rate by donor and recipient deme size using the equation m = (N_donor / N_recipient) * m_global. Default is True.",
    )
    demography_parser.add_argument(
        "-a",
        "--anc_pop_id",
        type=lambda x: x.split(",") if "," in x else x,
        default=None,
        help="List of ancestral population IDs or path to CSV file with column 'anc_pop_id'. Both have to be the same length as --coords. Only supply if you want to add ancestral populations. Default is None.",
    )
    demography_parser.add_argument(
        "-ts",
        "--timesteps",
        type=int,
        default=1,
        help="The list of timesteps representing the amount of time passing between each demographic event, in generations. If a single integer is provided, the function assumes that the time steps are equal. Default is 1.",
    )
    demography_parser.add_argument(
        "-as",
        "--anc_sizes",
        nargs="+",
        type=lambda x: [int(i) for i in x.split(",")],
        default=None,
        help="List of sizes for ancestral populations. Accepts a list of single values or a list of pairs of values. Default is None.",
    )
    demography_parser.add_argument(
        "-mt",
        "--merge_time",
        nargs="+",
        type=int,
        default=None,
        help="Time that demes merge into one or more ancestral populations. Measured in generations. Accepts a single int or a pair of ints. Default is None.",
    )
    demography_parser.add_argument(
        "-amt",
        "--anc_merge_time",
        nargs="+",
        type=int,
        default=None,
        help="Merge time for ancestral populations. Measured in generations. Accepts a single int or a pair of ints. Default is None.",
    )
    demography_parser.add_argument(
        "-ams",
        "--anc_merge_size",
        nargs="+",
        type=int,
        default=None,
        help="Merge size for ancestral populations. Accepts a single int or a pair of ints. Default is None.",
    )
    demography_parser.add_argument(
        "-amr",
        "--anc_mig_rate",
        nargs="+",
        type=float,
        default=None,
        help="Migration rate between ancestral populations. Accepts a single int or a pair of ints. Default is None.",
    )
    # Simulation Setup
    simulation_parser = parser.add_argument_group("Simulation Setup")
    simulation_parser.add_argument(
        "-sl",
        "--seq_length",
        type=int,
        default=1e6,
        help="Length of sequence to simulate. Measured in base pairs. Default is 1e6.",
    )
    simulation_parser.add_argument(
        "-mu",
        "--mutation_rate",
        nargs="+",
        type=float,
        default=[1e-8],
        help="Mutation rate per generation per base pair. Accepts a single float or a pair of floats. Default is [1e-8].",
    )
    simulation_parser.add_argument(
        "-rr",
        "--recombination_rate",
        nargs="+",
        type=float,
        default=[0],
        help="Recombination rate per generation per base pair. Accepts a single float or a pair of floats. Default is [0].",
    )
    simulation_parser.add_argument(
        "-pl",
        "--ploidy",
        type=int,
        default=2,
        help="Ploidy of individuals. Default is 2.",
    )
    simulation_parser.add_argument(
        "-npc",
        "--num_param_combos",
        type=int,
        default=1,
        help="Number of parameter combinations to simulate. Default is 1.",
    )
    simulation_parser.add_argument(
        "-ncs",
        "--num_coalescent_sims",
        type=int,
        default=1,
        help="Number of coalescent simulations to run per parameter combination. Default is 1.",
    )
    # Analysis Setup
    analysis_parser = parser.add_argument_group("Analysis Setup")
    analysis_parser.add_argument(
        "-mdp",
        "--missing_data_perc",
        type=float,
        default=0,
        help="Percentage of data masked as missing. Accepts a single float between 0 and 1. Default is 0.",
    )
    analysis_parser.add_argument(
        "-rt",
        "--r2_thresh",
        type=float,
        default=0.1,
        help="Threshold value for filtering sites based on linkage disequilibrium R^2. Default is 0.1.",
    )
    analysis_parser.add_argument(
        "-fm",
        "--filter_monomorphic",
        type=bool,
        default=True,
        help="Filter out monomorphic sites. Default is True.",
    )
    analysis_parser.add_argument(
        "-fs",
        "--filter_singletons",
        type=bool,
        default=True,
        help="Filter out singleton sites. Default is True.",
    )
    analysis_parser.add_argument(
        "-ss",
        "--sumstats",
        nargs="+",
        type=lambda x: x.split(",") if "," in x else x,
        default="all",
        help="List of summary statistics to calculate. Options are 'pi', 'tajima_d', 'sfs_h', 'fst', 'dxy', 'ibd', 'all'. For explanation of summary statistics, see the documentation. Default is all.",
    )
    analysis_parser.add_argument(
        "-wap",
        "--within_anc_pop_sumstats",
        type=bool,
        default=False,
        help="Whether to calculate summary statistics within ancestral populations. Defaults to False.",
    )
    analysis_parser.add_argument(
        "-bap",
        "--between_anc_pop_sumstats",
        type=bool,
        default=False,
        help="Whether to calculate Fst and/or Dxy between ancestral populations. Defaults to False.",
    )
    # Output
    parser.add_argument(
        "-ot",
        "--out_type",
        type=int,
        default=3,
        help="Type of output. 0 for writing tree sequences to file, 1 for writing VCFs to file, 2 for a CSV of genetic summary statistics, or 3 for writing all outputs to files. Default is 3.",
    )

    parser.add_argument(
        "-map",
        "--map",
        type=bool,
        default=False,
        help="Simulate the genetic diversity for each deme across entire landscape and output a GeoTiff. Overrides out_type. Default is False.",
    )

    parser.add_argument(
        "-of",
        "--out_folder",
        type=str,
        default=None,
        help="Path to output folder. Default is the current working directory.",
    )
    parser.add_argument(
        "-op",
        "--out_prefix",
        type=str,
        default="spaceprime",
        help="Prefix for output files. Default is 'spaceprime'.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=bool,
        default=False,
        help="Print progress to console. Default is False.",
    )
    parser.add_argument(
        "-c",
        "--cpu",
        type=int,
        default=1,
        help="Number of CPUs to use for parallel processing. Default is 1.",
    )

    args = parser.parse_args()

    # Check if params YAML file exists. If so, update command line arguments with parameters from YAML file
    if args.params is not None:
        if not os.path.exists(args.params):
            raise FileNotFoundError(f"Parameters file {args.params} not found")
        else:
            with open(args.params, "r") as file:
                params = yaml.safe_load(file)
            # Update command line arguments with parameters from YAML file
            for key, value in params.items():
                setattr(args, key, value)

    args.max_local_size = check_list_argument(args.max_local_size)
    args.inflection_point = check_list_argument(args.inflection_point)
    args.slope = check_list_argument(args.slope)
    args.mig_rate = check_list_argument(args.mig_rate)
    args.anc_pop_id = check_list_argument(args.anc_pop_id)
    args.anc_sizes = check_list_argument(args.anc_sizes)
    # make sure anc_sizes is a list of single value or a list of lists with two values in each element
    if args.anc_sizes[0] is not None:
        if isinstance(args.anc_sizes[0], list):
            args.anc_sizes = [list(map(int, size)) for size in args.anc_sizes]
            for size in args.anc_sizes:
                if len(size) != 2:
                    raise ValueError(
                        f"Each element in --anc_sizes must have two values"
                    )
        else:
            args.anc_sizes = [int(size) for size in args.anc_sizes]

    args.anc_merge_time = check_list_argument(args.anc_merge_time)
    args.anc_merge_size = check_list_argument(args.anc_merge_size)
    args.anc_mig_rate = check_list_argument(args.anc_mig_rate)

    # Check if list arguments that are supposed to be priors have more than two elements
    check_argument_length(args.max_local_size, 2)
    check_argument_length(args.inflection_point, 2)
    check_argument_length(args.slope, 2)
    check_argument_length(args.mig_rate, 2)
    check_argument_length(args.anc_merge_time, 2)
    check_argument_length(args.anc_merge_size, 2)
    check_argument_length(args.anc_mig_rate, 2)

    # check if raster file exists
    if not os.path.exists(args.raster):
        raise FileNotFoundError(f"Raster file {args.raster} not found")

    # check if out_folder exists
    if args.out_folder is not None:
        if not os.path.exists(args.out_folder):
            raise FileNotFoundError(f"Output folder {args.out_folder} not found")

    # generate parameter combinations
    param_combos = generate_param_combinations(args)
    logging.info("Generated parameter combinations")

    # run simulations in parallel
    if args.cpu == 1:
        for combo in param_combos:
            run_simulation(combo, args)
    else:
        with Pool(args.cpu) as p:
            p.starmap(run_simulation, [(combo, args) for combo in param_combos])



"""
Module for creating and modifying msprime Demography objects in the service of creating two-dimensional stepping-stone models.
"""

import numpy as np
from typing import Union, List, Optional
import msprime
from .utilities import calc_migration_matrix
import warnings


def add_landscape_change(
    model: msprime.Demography,
    d: np.ndarray,
    timesteps: Union[int, List[int]],
    rate: Union[float, np.ndarray] = 0.001,
    scale: bool = True,
) -> msprime.Demography:
    """
    If there are multiple time steps of deme size change, such as transformed species distribution model projections to past time periods, this function adds the changes to the model.
    It updates the deme sizes and migration rates at each time step.

    Parameters
    ----------
    model : msprime.Demography
        The model object to which the landscape change will be added.
    d : np.ndarray
        The 3D array representing different time steps of population size change.
    timesteps : Union[int, List[int]]
        The list of timesteps representing the amount of time passing between each demographic event, in generations. If a single integer is provided, the function assumes that the time steps are equal.
    rate : Union[float, np.ndarray], optional
        The migration rate. Defaults to 0.001.
    scale : bool, optional
        Whether to scale the migration rate based on population size. Defaults to True.

    Returns
    -------
    msprime.Demography
        The updated model object.
    """

    # check if the number of timesteps matches the number of layers in d, if timesteps is a list
    if isinstance(timesteps, list):
        assert (
            len(timesteps) == d.shape[0] - 1
        ), "Number of timesteps should match the number of layers in d - 1"

    # create a list to store each calculated migration matrix if the user provides a scale for migration rate
    mig_array = []

    # have to set the migration rate differently for the most recent time step, since it's specified prior to add_landscape_change
    if np.array(rate).ndim == 0:
        mig_array.append(calc_migration_matrix(d[0], rate, scale=scale))
    # if the user provides a 3D array of migration matrices, get the migration matrix for the current time step
    else:
        mig_array.append(rate[0])

    # iterate through the first dimension of a 3D array, where the array represents different time steps of population size change
    # omit the most ancient time step (have to set its migration rate differently)
    for step in range(1, d.shape[0]):
        # get the population size values of the current array
        kmat = d[step]

        # get the population size values of array from the more ancient time step
        if step < d.shape[0] - 1:
            kmat_anc = d[step + 1]
        else:
            kmat_anc = None

        # get the population size values array from the more recent time step
        kmat_prev = d[step - 1]

        # get the shape of the array
        n, m = kmat.shape

        # get the time for the demographic event
        if isinstance(timesteps, int):
            demo_time = timesteps * step
        else:
            demo_time = sum(timesteps[:step])

        # if the user provides a scalar migration rate, calculate the migration matrices for all time steps
        if np.array(rate).ndim == 0:
            migmat = calc_migration_matrix(kmat, rate, scale=scale)
        # if the user provides a 3D array of migration matrices, get the migration matrix for the current time step
        else:
            migmat = rate[step]

        # append the migration matrix to the list
        mig_array.append(migmat)

        ##### Update population sizes #####
        # add population size changes according to the values of the current array
        for j in range(n):
            for k in range(m):
                # only update the population size if it is different from the previous time point
                if kmat[j, k] != kmat_prev[j, k]:
                    # add a demographic change to each cell in the raster
                    model.add_population_parameters_change(
                        time=demo_time,
                        population=f"deme_{j}_{k}",
                        initial_size=kmat[j, k],
                    )

        ##### Update migration rates #####
        # add migration rate change for each time step
        # this is updating time steps from the present to the past

        for i in range(n):
            for j in range(m):
                for di, dj in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                    # check for edges
                    if 0 <= i + di < kmat.shape[0] and 0 <= j + dj < kmat.shape[1]:
                        ## only update migration if the donor and recipient population sizes are different between time points
                        if (
                            kmat_prev[i + di, j + dj] != kmat[i + di, j + dj]
                            and kmat[i, j] != kmat_prev[i, j]
                        ):
                            # get the migration rate from the migration matrix
                            rate = migmat[i + di, j + dj]
                            model.add_migration_rate_change(
                                time=demo_time,
                                rate=rate,
                                source=f"deme_{i}_{j}",
                                dest=f"deme_{i + di}_{j + dj}",
                            )
                        ## have the deme migrate to neighbors if the more ancient time step has an empty deme

                        elif (
                            ## only do this if kmat_anc is not none
                            kmat_anc is not None
                            and kmat_prev[i + di, j + dj] != kmat[i + di, j + dj]
                            and kmat[i, j] != kmat_prev[i, j]
                            and kmat_anc[i, j] <= 1e-9
                        ):
                            # get the migration rate from the migration matrix
                            rate = migmat[i + di, j + dj]
                            model.add_migration_rate_change(
                                time=demo_time,
                                rate=rate,
                                source=f"deme_{i}_{j}",
                                dest=f"deme_{i + di}_{j + dj}",
                            )

    # convert the list of migration matrices to a numpy array
    mig_array = np.array(mig_array)

    # sort the events to make sure everything runs in order
    model.sort_events()

    return model, mig_array


class spDemography(msprime.Demography):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def stepping_stone_2d(
        self,
        d: np.ndarray,
        rate: Union[float, np.ndarray],
        scale: bool = True,
        timesteps: Union[int, List[int]] = None,
    ) -> msprime.Demography:
        """
        Create a 2D stepping stone model, either for a single time step or for multiple time steps of deme size change.

        Parameters
        ----------
        d : numpy.ndarray
            The demography matrix representing the population sizes.
        rate : float or numpy.ndarray
            The migration rate(s) between populations.
            If a float, it represents a constant migration rate for all populations.
            If a numpy.ndarray, it represents a migration matrix with shape (T, N, N),
            where N is the total number of populations and T is the number of time steps - 1, if T > 1.
        scale : bool, optional
            Whether to scale the migration rate matrix. Default is True.
        timesteps : Union[int, List[int]], optional
            The list of timesteps representing the amount of time passing between each demographic event, in generations.
            If a single integer is provided, the function assumes that the time steps are equal.

        Returns
        -------
        model : msprime.Demography
            The constructed 2d stepping stone model as an msprime.Demography object.

        Notes
        -----
        The demography matrix `d` should have shape (n, m) or (k, n, m), where n is the number of rows and m is the number of columns for a 2D array and k is the number of layers in a 3D array.
        The migration rate matrix `rate` should have shape (N, N), where N is the total number of populations.
        If there are multiple time steps of population size change, the `add_landscape_change` function is called to modify the model accordingly.
        """

        if len(d.shape) == 2:
            n = d.shape[0]
            m = d.shape[1]
            N = n * m
            d1 = d  # so I don't have to repeat a bunch of code for the 3D case
        elif len(d.shape) == 3:
            t = d.shape[0]
            n = d.shape[1]
            m = d.shape[2]
            N = n * m
            d1 = d[0]  # so I don't have to repeat a bunch of code for the 3D case

        model = msprime.Demography.isolated_model(d1.reshape(N))

        # set population names
        for j in range(n):
            for k in range(m):
                index = j * m + k
                model.populations[index].name = f"deme_{j}_{k}"

        # setup migration rate matrices and check if the user provided their own migration matrix(es)
        if np.array(rate).ndim == 0:
            if scale:
                model.migration_matrix = calc_migration_matrix(d1, rate, scale=True)
            else:
                model.migration_matrix = calc_migration_matrix(d1, rate, scale=False)
        else:
            if len(rate.shape) == 2:
                assert rate.shape == (
                    N,
                    N,
                ), f"Expected a migration matrix with the shape {(N, N)} and instead got {rate.shape}"
                model.migration_matrix = rate
            else:
                # if the user supplies a 3D array of migration matrices
                assert rate.shape == (
                    t,
                    N,
                    N,
                ), f"Expected a migration matrix with the shape {(t, N, N)} and instead got {rate.shape}"
                model.migration_matrix = rate[0]

        # if there are multiple time steps of population size change
        if len(d.shape) == 3:
            model, mig_array = add_landscape_change(model, d, timesteps, rate, scale)
        else:
            mig_array = model.migration_matrix

        model.sort_events()

        # set the model attributes to the spDemography object
        self.__dict__.update(model.__dict__)
        # return a 3d migration array for each time step
        self.migration_array = mig_array
        # add demes to the object
        self.demes = d

    def add_ancestral_populations(
        self,
        anc_sizes: List[float],
        merge_time: Union[float, int],
        anc_id: Optional[np.ndarray] = None,
        anc_merge_times: Optional[List[float]] = None,
        anc_merge_sizes: Optional[List[float]] = None,
        migration_rate: Optional[float] = None,
    ) -> msprime.Demography:
        """
        Adds ancestral populations to the given demographic model, mapping demes in the spatial simulation to ancestral populations.

        Parameters
        ----------
        anc_sizes : List[float]
            A list of ancestral population sizes.
        merge_time : Union[float, int]
            The time at which all demes in the spatial simulation merge into one or more ancestral populations.
        anc_id : Optional[np.ndarray], optional
            An array of ancestral population IDs- the output of [split_landscape_by_pop][utilities.split_landscape_by_pop]. Defaults to None.
        anc_merge_times : Optional[List[float]], optional
            A list of merge times for ancestral populations. Defaults to None.
        anc_merge_sizes : Optional[List[float]], optional
            A list of sizes for merged ancestral populations. Defaults to None.
        migration_rate : Optional[float], optional
            The symmetric migration rate between ancestral populations. Defaults to None.

        Returns
        -------
        msprime.Demography
            The demographic model with the added ancestral populations.

        Raises
        ------
        ValueError
            If the model already contains ancestral populations.
        ValueError
            If the number of demes in the demographic model does not match the number of demes in the admixture ID raster.

        Notes
        -----
        The function adds ancestral populations to the given demographic model. If `anc_id` is not provided, a single ancestral population is added with the initial size specified in `anc_sizes[0]`. If `anc_id` is provided, a new ancestral population is added for each admixture population, with sizes specified in `anc_sizes`. The demes in the simulation are then merged into their respective ancestral populations based on the values in `anc_id`. If `anc_merge_times` is provided, the ancestral populations are merged at the specified times. If `migration_rate` is provided, symmetric migration is allowed between ancestral populations.
        """

        if any("ANC" in pop.name for pop in self.populations):
            raise ValueError(
                "Model already contains ancestral populations. You need to re-initialize the 2D stepping stone model before adding ancestral populations."
            )

        # convert anc_sizes to a list if it is a single value
        if anc_sizes is not None and not isinstance(anc_sizes, list):
            anc_sizes = [anc_sizes]

        # set anc_id to None and throw a warning if anc_sizes is less than 2
        if len(anc_sizes) < 2 and anc_id is not None:
            warnings.warn(
                "anc_sizes is less than two, but you provided a list of anc_id's. If you meant to setup more than one anc_size, provide a list of sizes to anc_size. Otherwise, you're fine."
            )
            anc_id = None

        # check merge time
        if isinstance(merge_time, list):
            raise ValueError("merge_time should be a single float or int, not a list.")
        elif not isinstance(merge_time, (float, int)):
            raise TypeError("merge_time should be a float or int.")

        # convert anc_merge_times to a list if it is a single value
        if anc_merge_times is not None and not isinstance(anc_merge_times, list):
            anc_merge_times = [anc_merge_times]

        # convert anc_merge_sizes to a list if it is a single value
        if anc_merge_sizes is not None and not isinstance(anc_merge_sizes, list):
            anc_merge_sizes = [anc_merge_sizes]

        if anc_id is None:
            # add an ancestral population
            self.add_population(name="ANC_1", initial_size=anc_sizes[0])

            # get names of populations for initiating the merge
            pop_names = [
                [pop.name] for pop in self.populations if "ANC" not in pop.name
            ]

            # add the time when the spatial simulation collapses into the collecting phase
            [
                self.add_population_split(
                    time=merge_time, derived=name, ancestral="ANC_1"
                )
                for name in pop_names
            ]
        else:
            # make sure the ancestral population ID array has the same number of populations as the array used in the demographic modeling
            try:
                anc_id_1d = anc_id.ravel()
                if len(self.populations) != len(anc_id_1d):
                    raise ValueError
            except ValueError:
                print(
                    f"The number of demes in the demographic model is {len(self.populations)}, while the number of demes in the admixture ID raster is {len(anc_id_1d)}. They need to be the same."
                )

            # add a new ancestral population
            for i in range(1, len(anc_sizes) + 1):
                anc_pop_name = f"ANC_{i}"
                self.add_population(name=anc_pop_name, initial_size=anc_sizes[i - 1])

            # merge each deme in the simulation into its respective ancestral population
            for i in range(len(self.populations)):
                pop_name = self.populations[i].name
                if "ANC" not in pop_name:
                    anc_pop = f"ANC_{int(anc_id_1d[i])}"
                    self.add_population_split(
                        time=merge_time, derived=[pop_name], ancestral=anc_pop
                    )

            # merge ancestral populations at their respective times, if we want that behavior
            if anc_merge_times is not None:
                try:
                    if (
                        len(anc_merge_times) != len(anc_sizes) + 1
                        and len(anc_merge_times) != 1
                    ):
                        raise ValueError
                except ValueError:
                    print("The ancestral merge list should be of length N - 1 or 1")

                # make sure the ancestral merge list is either of size 1 or N - 1
                if len(anc_sizes) > 1 and len(anc_merge_times) > 1:
                    for i in range(1, len(anc_sizes)):
                        if i == 1:
                            n1 = f"ANC_{i}"
                            n2 = f"ANC_{i + 1}"
                            anc_n = f"ANC_{i}_{i + 1}"
                        else:
                            anc_num_str = "_".join(map(str, range(1, i + 1)))
                            n1 = f"ANC_{anc_num_str}"
                            n2 = f"ANC_{i + 1}"
                            anc_n = f"ANC_{anc_num_str}_{i + 1}"
                        ## add the new ancestral populations to the simulation
                        self.add_population(
                            name=anc_n, initial_size=anc_merge_sizes[i - 1]
                        )
                        self.add_population_split(
                            time=anc_merge_times[i - 1],
                            derived=[n1, n2],
                            ancestral=anc_n,
                        )
                ## if there are multiple ancestral populations, but a single merge time
                elif len(anc_sizes) > 1 and len(anc_merge_times) == 1:
                    # get a list of the ancestral population names
                    anc_der_pops = []
                    for i in range(1, len(anc_sizes) + 1):
                        anc_der_name = f"ANC_{i}"
                        anc_der_pops.append(anc_der_name)
                    # make the name of the most ancestral population
                    anc_num_str = "_".join(map(str, range(1, len(anc_sizes) + 1)))
                    anc_n = f"ANC_{anc_num_str}"
                    ## add the new ancestral populations to the simulation
                    self.add_population(name=anc_n, initial_size=anc_merge_sizes[0])
                    self.add_population_split(
                        time=anc_merge_times[0], derived=anc_der_pops, ancestral=anc_n
                    )

            # add migration between ancestral populations
            if migration_rate is not None:
                for i in range(1, len(anc_sizes) + 1):
                    for j in range(i + 1, len(anc_sizes) + 1):
                        self.set_symmetric_migration_rate(
                            populations=[f"ANC_{i}", f"ANC_{j}"],
                            rate=migration_rate,
                        )

        # sort the events to make sure everything runs in order
        self.sort_events()



"""
Module providing thin wrappers around msprime.sim_ancestry and msprime.sim_mutations.
"""

import msprime


def sim_ancestry(*args, **kwargs):
    """
    A thin wrapper around msprime.sim_ancestry.

    Parameters
    ----------
    *args : tuple
        Variable length argument list.
    **kwargs : dict
        Arbitrary keyword arguments.

    Returns
    -------
    Union[TreeSequence, Iterator[TreeSequence]]
        The TreeSequence object representing the results of the simulation if no replication is performed,
        or an iterator over the independent replicates simulated if the num_replicates parameter has been used.

    Notes
    -----
    This function takes the same arguments as msprime.sim_ancestry and calls it directly,
    allowing users to use simulation functionality within the spaceprime namespace.

    See the msprime.sim_ancestry documentation for more information: See the msprime documentation for more information: https://tskit.dev/msprime/docs/stable/api.html#msprime.sim_ancestry

    Returns
    -------
    Union[TreeSequence, Iterator[TreeSequence]]
        The TreeSequence object representing the results of the simulation if no replication is performed,
        or an iterator over the independent replicates simulated if the num_replicates parameter has been used.
    """
    return msprime.sim_ancestry(*args, **kwargs)


def sim_mutations(*args, **kwargs):
    """
    A thin wrapper around msprime.sim_mutations.

    Parameters
    ----------
    *args : Variable length argument list.
    **kwargs : Arbitrary keyword arguments.

    Returns
    -------
    The result of msprime.sim_mutations with the provided arguments.

    Notes
    -----
    This function takes the same arguments as msprime.sim_ancestry and calls it directly,
    allowing users to use simulation functionality within the spaceprime namespace.

    See the msprime.sim_mutations documentation for more information: https://tskit.dev/msprime/docs/stable/api.html#msprime.sim_mutations

    """
    return msprime.sim_mutations(*args, **kwargs)



#!/usr/bin/env python

"""Tests for `plot` module."""
import pytest
import numpy as np
from spaceprime import plot
from spaceprime import demography
import rasterio
from math import isclose
import matplotlib
from matplotlib import pyplot as plt
import folium


@pytest.fixture
def demes():
    return np.array([[1000, 500], [200, 300]])


@pytest.fixture
def mig_mat():
    # 0.01 is the migration rate between demes with scaling
    return np.array(
        [
            [0.0, 0.005, 0.002, 0.0],
            [0.02, 0.0, 0.0, 0.006],
            [0.05, 0.0, 0.0, 0.015],
            [0.0, 0.01666667, 0.00666667, 0.0],
        ]
    )


@pytest.fixture
def demo(demes, mig_mat):
    d = demography.spDemography()
    # Add demes to the demography object
    d.add_population(name="deme1", initial_size=demes[0][0])
    d.add_population(name="deme2", initial_size=demes[0][1])
    d.add_population(name="deme3", initial_size=demes[1][0])
    d.add_population(name="deme4", initial_size=demes[1][1])
    # Add migration rates between demes
    d.set_migration_rate("deme1", "deme2", rate=mig_mat[0][1])
    d.set_migration_rate("deme1", "deme3", rate=mig_mat[0][2])
    d.set_migration_rate("deme1", "deme4", rate=mig_mat[0][3])
    d.set_migration_rate("deme2", "deme1", rate=mig_mat[1][0])
    d.set_migration_rate("deme2", "deme3", rate=mig_mat[1][2])
    d.set_migration_rate("deme2", "deme4", rate=mig_mat[1][3])
    d.set_migration_rate("deme3", "deme1", rate=mig_mat[2][0])
    d.set_migration_rate("deme3", "deme2", rate=mig_mat[2][1])
    d.set_migration_rate("deme3", "deme4", rate=mig_mat[2][3])
    d.set_migration_rate("deme4", "deme1", rate=mig_mat[3][0])
    d.set_migration_rate("deme4", "deme2", rate=mig_mat[3][1])
    d.set_migration_rate("deme4", "deme3", rate=mig_mat[3][2])

    # add demes and migration_array attributes to the spDemography object
    d.demes = []
    d.demes.append(demes)

    # add a second timestep to the demes array
    d.demes.append(demes * 2)

    d.migration_array = []
    d.migration_array.append(mig_mat)

    return d


@pytest.fixture
def raster():
    return rasterio.open("tests/data/demes_raster.tif")


# get_outgoing_migration_rates
def test_get_outgoing_migration_rates_values(demes, mig_mat):
    outgoing_migration_rates = plot.get_outgoing_migration_rates(demes, mig_mat)
    assert outgoing_migration_rates.shape == (4, 5)
    assert outgoing_migration_rates["deme_size"].sum() == 2000
    assert isclose(
        outgoing_migration_rates["mig_north"].sum(), 0.066666667, rel_tol=1e-5
    )
    assert outgoing_migration_rates["mig_east"].sum() == 0.02
    assert outgoing_migration_rates["mig_south"].sum() == 0.008
    assert isclose(outgoing_migration_rates["mig_west"].sum(), 0.02666666, rel_tol=1e-5)


def test_get_outgoing_migration_rates_navalue(demes, mig_mat):
    """Test that get_outgoing_migration_rates skips NA values in the migration matrix"""

    omr = plot.get_outgoing_migration_rates(demes, mig_mat, na_value=200)

    assert omr.shape == (3, 5)


# plot_model
def test_plot_model_timestep(demo, raster) -> None:
    """Test that the plot_model function raises a ValueError if the timestep is invalid"""

    with pytest.raises(ValueError):
        plot.plot_model(demo, raster, 2)


def test_plot_model_demo(demo, raster) -> None:
    """Test that the plot_model function raises a ValueError if the demography object isn't populated with deme sizes yet"""

    d1 = demography.spDemography()
    with pytest.raises(ValueError):
        plot.plot_model(d1, raster, 0)


def test_plot_model_returns_plot(demo, raster) -> None:
    # Call the function
    p = plot.plot_model(demo, raster, 0)

    # Test 1: Check if a folium.Map object is returned
    assert isinstance(p, folium.Map)


# plot_landscape
def test_plot_landscape_demo(demo, raster) -> None:
    """Test that the plot_landscape function raises a ValueError if the demography object isn't populated with deme sizes yet"""

    d1 = demography.spDemography()
    with pytest.raises(ValueError):
        plot.plot_landscape(d1, raster, 0)


def test_plot_landscape_timestep(demo, raster) -> None:
    """Test that the plot_landscape function raises a ValueError if the timestep is invalid"""

    with pytest.raises(ValueError):
        plot.plot_landscape(demo, raster, 3)


def test_plot_landscape_returns_plot(demo, raster) -> None:
    """Test that the plot_landscape function returns a matplotlib axes object"""

    p = plot.plot_landscape(demo, raster, 0)
    assert isinstance(p, matplotlib.axes.Axes)


# plot_timeseries
def test_plot_timeseries_data(demo):
    """Test that the plot_timeseries function returns the correct data"""

    times = [0, 1]

    fig, ax = plot.plot_timeseries(demo, times, "Generations")
    x_plot, y_plot = ax.lines[0].get_xydata().T

    # get the y values
    total_individuals = [np.nansum(d) for d in demo.demes]

    assert np.all(x_plot == times)
    assert np.all(y_plot == total_individuals)


def test_plot_timeseries_times_length_1(demo):
    """Test that the plot_timeseries function raises a ValueError if the times list is less than 2"""

    times = [0]

    with pytest.raises(ValueError):
        plot.plot_timeseries(demo, times, "Generations")


def test_plot_timeseries_times_length_not_equal_to_timesteps(demo):
    """Test that the plot_timeseries function raises a ValueError if the length of the times list is not equal to the number of timesteps"""

    times = [0, 1, 2]

    with pytest.raises(ValueError):
        plot.plot_timeseries(demo, times, "Generations")
